"""
Maxar VHR stereo imagery

https://www.maxar.com/maxar-intelligence/products/satellite-imagery
"""

# from pydantic.dataclasses import dataclass, Field # type: ignore[attr-defined]
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from stac_asset import Config

import os
import warnings
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path

import numpy as np
import pystac
import rasterio
import stac_asset
import xarray as xr

from coincident._utils import depends_on_optional
from coincident.datasets.general import Dataset

MAXAR_CONFIG = stac_asset.Config(
    http_headers={"MAXAR-API-KEY": os.environ.get("MAXAR_API_KEY")}
)
try:
    import matplotlib.pyplot as plt
except ImportError:
    warnings.warn(
        "'matplotlib' package not found. Install for plotting functions: https://matplotlib.org/stable/install/index.html",
        stacklevel=2,
    )


class Collection(str, Enum):
    """wv01,wv02,wv03-vnir,ge01"""

    wv01 = "wv01"
    wv02 = "wv02"
    wv03_vnir = "wv03-vnir"
    ge01 = "ge01"


@dataclass
class Stereo(Dataset):
    """Essential metadata for Maxar In Track Stereo"""

    alias: str = "maxar"
    has_stac_api: bool = True
    collections: list[Collection] = field(
        default_factory=lambda: ["wv01", "wv02", "wv03-vnir", "ge01"]
    )  # type: ignore[assignment]
    search: str = "https://api.maxar.com/discovery/v1/search"
    start: str = "2007-01-01"
    end: str | None = None
    type: str = "stereo"
    # Unique to Maxar
    area_based_calc: bool = False
    provider: str = "maxar"


# NOTE: expose as coincident.search.download
# Or just add documentation for how to do this?
async def download_item(
    item: pystac.Item,
    path: str = "/tmp",
    config: Config = MAXAR_CONFIG,
) -> pystac.Item:
    """localitem = asyncio.run(download_item(item, config=MAXAR_CONFIG))"""
    posixpath = Path(path)
    item = await stac_asset.download_item(item, posixpath, config=config)
    return item  # noqa: RET504


def open_browse(item: pystac.Item, overview_level: int = 0) -> xr.DataArray:
    """
    Open a browse image from a STAC item using the specified overview level.

    Parameters
    ----------
    item : pystac.Item
        The STAC item containing the browse image asset.
    overview_level : int, optional
        The overview level to use when opening the image, by default 0.

    Returns
    -------
    xr.DataArray
        The opened browse image as an xarray DataArray.

    Notes
    -----
    The function uses the `rasterio` engine to open the image and sets the
    `GDAL_DISABLE_READDIR_ON_OPEN` and `GDAL_HTTP_HEADERS` environment variables
    for optimized reading and authentication, respectively.
    """

    # href = item.assets['browse']['href'] # accessed via geopandas
    href = item.assets["browse"].href  # accessed via pystac

    env = rasterio.Env(
        GDAL_DISABLE_READDIR_ON_OPEN="EMPTY_DIR",
        GDAL_HTTP_HEADERS=f'MAXAR-API-KEY:{os.environ["MAXAR_API_KEY"]}',
    )
    with env:
        return xr.open_dataarray(
            href,
            engine="rasterio",
            mask_and_scale=False,  # otherwise uint8 -> float32!
            backend_kwargs={"open_kwargs": {"overview_level": overview_level}},
        )


@depends_on_optional("matplotlib")
def plot_browse(item: pystac.Item, overview_level: int = 0) -> None:
    """
    Plots a browse image from a STAC item using Matplotlib.

    Parameters
    ----------
    item : pystac.Item
        The STAC item containing the browse image to be plotted.
    overview_level : int, optional
        The overview level of the browse image to be opened, by default 0.

    Returns
    -------
    None
        This function does not return any value
    """

    da = open_browse(item, overview_level=overview_level)
    mid_lat = da.y[int(da.y.size / 2)].to_numpy()  # PD011

    fig, ax = plt.subplots(figsize=(8, 11))  # pylint: disable=unused-variable
    da.plot.imshow(rgb="band", add_labels=False, ax=ax)
    ax.set_aspect(aspect=1 / np.cos(np.deg2rad(mid_lat)))
    ax.set_title(item.id)



"""
Subset and Sample using Xarray
"""

from __future__ import annotations

import warnings
from typing import Any

import geopandas as gpd
import odc.stac

# NOTE: must import for odc.stac outputs to have .rio accessor
import rioxarray  # noqa: F401
import xarray as xr

from coincident._utils import depends_on_optional
from coincident.datasets.planetary_computer import WorldCover
from coincident.search.stac import to_pystac_items

try:
    import matplotlib.colors
    import matplotlib.pyplot as plt
    from matplotlib import cm
except ImportError:
    warnings.warn(
        "'matplotlib' package not found. Install for plotting functions: https://matplotlib.org/stable/install/index.html",
        stacklevel=2,
    )

# Sets GDAL_DISABLE_READDIR_ON_OPEN to 'EMPTY_DIR' etc.
odc.stac.configure_rio(cloud_defaults=True)


def to_dataset(
    gf: gpd.GeoDataFrame,
    bands: list[str] | None = None,
    aoi: gpd.GeoDataFrame | None = None,
    mask: bool = False,
    **kwargs: dict[str, Any],
) -> xr.DataArray:
    """
    Convert a GeoDataFrame to an xarray DataArray using odc.stac

    Parameters
    ----------
    gf : gpd.GeoDataFrame
        The GeoDataFrame containing the geospatial data.
    aoi : gpd.GeoDataFrame, optional
        GeoDataFrame containing area of interest (e.g. search polygon)
    bands : str, optional
        The asset keys to extract from the STAC items, by default ["data"].
    **kwargs : dict
        Additional keyword arguments passed to `odc.stac.load`.

    Returns
    -------
    xr.Dataset
        The resulting stacked xarray Dataset backed by dask arrays
    """
    if bands is None:
        bands = ["data"]

    items = to_pystac_items(gf)
    ds = odc.stac.load(items, bands=bands, geopolygon=aoi, chunks={}, **kwargs)
    if mask:
        ds = ds.rio.clip(aoi)

    return ds


@depends_on_optional("matplotlib")
def plot_esa_worldcover(ds: xr.Dataset) -> plt.Axes:
    """From https://planetarycomputer.microsoft.com/dataset/esa-worldcover#Example-Notebook"""
    classmap = WorldCover().classmap

    colors = ["#000000" for r in range(256)]
    for key, value in classmap.items():
        colors[int(key)] = value["hex"]
    cmap = matplotlib.colors.ListedColormap(colors)

    # sequences needed for an informative colorbar
    values = list(classmap)
    boundaries = [(values[i + 1] + values[i]) / 2 for i in range(len(values) - 1)]
    boundaries = [0, *boundaries, 255]
    ticks = [
        (boundaries[i + 1] + boundaries[i]) / 2 for i in range(len(boundaries) - 1)
    ]
    tick_labels = [value["description"] for value in classmap.values()]

    fig, ax = plt.subplots()
    normalizer = matplotlib.colors.Normalize(vmin=0, vmax=255)

    da = ds.to_dataarray().squeeze()
    da.plot(ax=ax, cmap=cmap, norm=normalizer)

    colorbar = fig.colorbar(
        cm.ScalarMappable(norm=normalizer, cmap=cmap),
        boundaries=boundaries,
        values=values,
        cax=fig.axes[1].axes,
    )
    colorbar.set_ticks(ticks, labels=tick_labels)
    return ax



from __future__ import annotations

import typing

import geopandas as gpd
import pytest
from geopandas.testing import assert_geodataframe_equal

import coincident

# Decorate tests requiring internet (slow & flaky)
network = pytest.mark.network

try:
    import maxar_platform.discovery  # noqa: F401

    not_authenticated = False
except:  # noqa: E722
    not_authenticated = True
maxar_authenticated = pytest.mark.skipif(
    not_authenticated, reason="tests for linux only"
)


@pytest.fixture
def aoi():
    # 11 vertices, 1,361km^2
    aoi_url = "https://raw.githubusercontent.com/SlideRuleEarth/sliderule-python/main/data/grandmesa.geojson"
    return gpd.read_file(aoi_url)


@pytest.fixture
def large_aoi():
    # 260 vertices, large area 269,590 km^2
    aoi_url = "https://raw.githubusercontent.com/unitedstates/districts/refs/heads/gh-pages/states/CO/shape.geojson"
    return gpd.read_file(aoi_url)


@typing.no_type_check
def test_no_dataset_specified():
    with pytest.raises(
        TypeError, match="missing 1 required positional argument: 'dataset'"
    ):
        coincident.search.search(intersects="-120, 40, -121, 41")


def test_unknown_dataset_specified():
    with pytest.raises(ValueError, match="is not a supported dataset"):
        coincident.search.search(dataset="typo", intersects="-120, 40, -121, 41")


def test_polygon_invalid_type():
    with pytest.raises(
        ValueError, match="intersects value must be a GeoDataFrame or GeoSeries"
    ):
        coincident.search.search(dataset="3dep", intersects="-120, 40, -121, 41")


def test_to_geopandas_empty_search_result():
    with pytest.raises(ValueError, match="ItemCollection is empty"):
        coincident.search.stac.to_geopandas([])


def test_unconstrained_search_warns():
    with pytest.warns(match="Neither `bbox` nor `intersects` provided"):
        coincident.search.search(dataset="tdx")


@network
def test_cascading_search(aoi):
    aoi["datetime"] = gpd.pd.to_datetime("2019-06-12", utc=True)
    pad = 30
    secondary_datasets = [("icesat-2", pad), ("gedi", pad)]
    results = coincident.search.cascading_search(aoi, secondary_datasets)

    expected_min = aoi.datetime.iloc[0] - gpd.pd.Timedelta(days=pad)
    expected_max = aoi.datetime.iloc[0] + gpd.pd.Timedelta(days=pad)
    actual_min = results[0].datetime.min()
    actual_max = results[0].datetime.max()

    assert isinstance(results, list)
    assert len(results) == 2
    assert len(results[0]) == 5
    assert len(results[1]) == 4
    assert actual_min >= expected_min
    assert actual_max <= expected_max


# TODO: add more assertions / tests for this section
@network
@maxar_authenticated
@pytest.mark.filterwarnings("ignore:Server does not conform")
def test_maxar_search(aoi):
    gf = coincident.search.search(
        dataset="maxar",
        intersects=aoi,
        datetime="2023",
        filter="eo:cloud_cover < 20",
    )
    assert len(gf) == 10
    assert isinstance(gf.datetime.iloc[0], gpd.pd.Timestamp)
    # NOTE: add more comprehensive check of column names & schema?
    assert "id" in gf.columns
    assert "assets" in gf.columns


@network
@maxar_authenticated
def test_maxar_large_aoi(large_aoi):
    gf = coincident.search.search(
        dataset="maxar",
        intersects=large_aoi,
        datetime="2023",
        # NOTE: this limits total *mono* results, which may or may not have stereo
        max_items=10,
    )
    assert len(gf) <= 10


# NASA
# =======
@network
def test_icesat2_search(aoi):
    gf = coincident.search.search(
        dataset="icesat-2",
        intersects=aoi,
        datetime="2023",
    )
    assert len(gf) == 25
    assert isinstance(gf.start_datetime.iloc[0], gpd.pd.Timestamp)


@network
def test_gedi_search(aoi):
    gf = coincident.search.search(
        dataset="gedi",
        intersects=aoi,
        datetime="2022",
    )
    assert len(gf) == 33


@network
def test_tdx_search(aoi):
    gf = coincident.search.search(
        dataset="tdx", intersects=aoi, datetime=["2009", "2020"]
    )
    assert len(gf) == 48
    assert gf["sar:product_type"].unique() == "SSC"


# MS PLANETARY COMPUTER
# =======
@network
def test_cop30_search(aoi):
    gf = coincident.search.search(dataset="cop30", intersects=aoi)
    assert len(gf) == 4


@network
def test_worldcover_search(aoi):
    gf = coincident.search.search(dataset="worldcover", intersects=aoi, datetime="2020")
    assert len(gf) == 4


@network
def test_round_trip_parquet(aoi):
    outpath = "/tmp/search_results.parquet"
    A = coincident.search.search(dataset="cop30", intersects=aoi)
    A.to_parquet(outpath)
    B = gpd.read_parquet(outpath)
    assert_geodataframe_equal(A, B)


# USGS
# =======
@network
def test_wesm_search(aoi):
    gf = coincident.search.search(
        dataset="3dep",
        intersects=aoi,
    )
    assert len(gf) == 5


# NOTE ~10s on wifi
@network
def test_get_swath_polygon():
    gf = coincident.search.wesm.get_swath_polygons("CO_CameronPkFire_1_2021")
    assert isinstance(gf, gpd.GeoDataFrame)
    assert len(gf) == 51
    assert "start_datetime" in gf.columns
    assert isinstance(gf.datetime.iloc[0], gpd.pd.Timestamp)


@network
def test_swath_polygon_not_found():
    with pytest.raises(ValueError, match="No swath polygons found for workunit="):
        coincident.search.wesm.get_swath_polygons("AL_SWCentral_1_B22")



"""
NASA Datasets

https://github.com/nasa/cmr-stac

NOTE: Different search endpoints dependng on DAAC
"""

# from pydantic.dataclasses import dataclass
from __future__ import annotations

from dataclasses import dataclass, field

from coincident.datasets.general import Dataset


@dataclass
class ICESat2(Dataset):
    """Essential metadata for ICESat-2 Alitmeter"""

    # https://cmr.earthdata.nasa.gov/stac/NSIDC_ECS/collections/ATL03_006
    has_stac_api: bool = True
    search: str = "https://cmr.earthdata.nasa.gov/stac/NSIDC_ECS"
    start: str | None = "2018-10-13"
    type: str = "altimeter"
    alias: str = "icesat-2"
    collections: list[str] = field(
        default_factory=lambda: ["ATL03_006"]
    )  # ATL08_006 etc.
    provider: str = "nasa"


@dataclass
class GEDI(Dataset):
    """Essential metadata for GEDI Altimeter"""

    # NOTE: parse temporal & bbox from collection metadata?
    # https://cmr.earthdata.nasa.gov/stac/LPCLOUD/collections/GEDI02_A_002
    has_stac_api: bool = True
    search: str = "https://cmr.earthdata.nasa.gov/stac/LPCLOUD"
    start: str = "2019-04-04"
    # https://www.earthdata.nasa.gov/news/nasa-announces-pause-gedi-mission
    end: str = "2023-03-17"
    type: str = "altimeter"
    alias: str = "gedi"
    collections: list[str] = field(default_factory=lambda: ["GEDI02_A_002"])
    provider: str = "nasa"
